#include "stdafx.h"
#include "SortedArray.h"
#include "SimpleArrayEx.h"

//The one and only ATL module instance
#pragma warning(suppress: 26426)
ATL::CComModule _Module;


//A Simple function we use for sorting the test arrays we will use
int CompareInts(const int& i1, const int& i2) noexcept
{
  if (i1 > i2)
    return 1;
  else if (i1 < i2)
    return -1;
  else
    return 0;
}

//A functor version of the CompareInts function above
class CompareIntsClass
{
public:
  int operator()(int i1, int i2) const noexcept
  {
    if (i1 > i2)
      return 1;
    else if (i1 < i2)
      return -1;
    else
      return 0;
  }
};


//A simple class which contains an integer which is what we will be 
//using for the comparison and another int which represents the original
//index prior to the array being sorted
class CStabilityTestItem
{
public:
//Constructors / Destructors
  CStabilityTestItem() noexcept : m_nValue{0},
                                  m_nOriginalIndex{-1}
  {
  }
  CStabilityTestItem(int nValue, INT_PTR nOriginalIndex) noexcept : m_nValue{nValue},
                                                                    m_nOriginalIndex{nOriginalIndex}
  {
  }
  CStabilityTestItem(const CStabilityTestItem&) = default;
  CStabilityTestItem(CStabilityTestItem&&) = default;
  ~CStabilityTestItem() = default;

//Methods
  CStabilityTestItem& operator=(CStabilityTestItem&& item) = default;
  CStabilityTestItem& operator=(const CStabilityTestItem& item) = default;

//Member variables
  int m_nValue;
  INT_PTR m_nOriginalIndex;
};


//A functor for comparing CStabilityTestItem's which also is stable
class CompareCStabilityTestItemsClass
{
public:
  int operator()(const CStabilityTestItem& item1, const CStabilityTestItem& item2) const noexcept
  {
    if (item1.m_nValue > item2.m_nValue)
      return 1;
    else if (item1.m_nValue < item2.m_nValue)
      return -1;
    else
      return 0;
  }
};



#pragma warning(suppress: 26485)
int _tmain(int /*argc*/, TCHAR* /*argv*/[], TCHAR* /*envp*/[])
{
#ifndef SORTEDARRAY_ATL_ONLY
  //initialize MFC and print and error on failure (Note we must put up with some /analyze warnings for AfxWinInit even in VC 2008)
  if (!AfxWinInit(::GetModuleHandle(nullptr), nullptr, ::GetCommandLine(), 0))
    return 1;
#endif //#ifndef SORTEDARRAY_ATL_ONLY

  _tprintf(_T("Test Sort method (values should appear in ascending order)\n"));
  CSortedArrayEx<int, CompareIntsClass, const int> arr6;
  arr6.Add(7);
  arr6.Add(4);
  arr6.Add(0);
  arr6.Add(4);
  arr6.Add(2);
  arr6.Sort();
  for (INT_PTR i{0}; i<arr6.GetSize(); i++)
    _tprintf(_T("%d "), arr6[i]);
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr6.IsSorted());

  //Test out the CSortedArray methods
  CSortedArray<int, const int&> idata;
  idata.SetCompareFunction(CompareInts);
  int d{7};
  INT_PTR nIndex{idata.FindOrderedInsertIndex(d)}; //NOLINT(clang-analyzer-deadcode.DeadStores)
  INT_PTR nIndex2{idata.OrderedInsert(d)}; //NOLINT(clang-analyzer-deadcode.DeadStores, clang-diagnostic-unused-but-set-variable)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 3;
  nIndex = idata.FindOrderedInsertIndex(d); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex2 = idata.OrderedInsert(d); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 4;
  nIndex = idata.FindOrderedInsertIndex(d); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex2 = idata.OrderedInsert(d); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 5;
  nIndex = idata.FindOrderedInsertIndex(d); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex2 = idata.OrderedInsert(d); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 2;
  nIndex = idata.FindOrderedInsertIndex(d); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex2 = idata.OrderedInsert(d); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 11;
  idata.Add(d);
  d = 1;
  idata.Add(d);
  d = 2;
  idata.Add(d);
  d = 6;
  idata.Add(d);
  d = 3;
  idata.Add(d);
  idata.Add(d);
  idata.Add(d);
  idata.Add(d);
  idata.Add(d);
  idata.Add(d);
  idata.Add(d);

  idata.Sort();
#pragma warning(suppress: 26477)
  ATLASSERT(idata.GetSize() == 16);
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(idata.IsSorted());
  d = 11;
  nIndex = idata.Find(d); //NOLINT(clang-analyzer-deadcode.DeadStores)

  d = 6;
  nIndex = idata.Find(d); //NOLINT(clang-analyzer-deadcode.DeadStores)
  d = 3;
  nIndex = idata.Find(d); //NOLINT(clang-analyzer-deadcode.DeadStores)
  d = 2;
  nIndex = idata.Find(d); //NOLINT(clang-analyzer-deadcode.DeadStores)

  d = 88;
  nIndex = idata.Find(d);
  UNREFERENCED_PARAMETER(nIndex);

  _tprintf(_T("\nItems in array prior to calling UniqueSort\n  "));
  for (INT_PTR i{0}; i<idata.GetSize(); i++)
    _tprintf(_T("%d "), idata[i]);

  idata.UniqueSort();
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(idata.IsSorted());
#pragma warning(suppress: 26477)
  ATLASSERT(idata.GetSize() == 8); //Should have removed 1 duplicate "2" and 8 duplicate "3"

  _tprintf(_T("\nItems in array after calling UniqueSort\n  "));
  for (INT_PTR i{0}; i<idata.GetSize(); i++)
    _tprintf(_T("%d "), idata[i]);



  //Lets test the stability of the StableSort method. We do this by storing the original index value each element
  //has in the array before we call "Sort" and then finally check all the index values are still sequential
  CSortedArrayEx<CStabilityTestItem, CompareCStabilityTestItemsClass, const CStabilityTestItem&> arr5;
  _tprintf(_T("\nTesting StableSort stability\n"));
  int nItemsToAdd{1000};
  arr5.SetSize(0, nItemsToAdd);
  for (int i{0}; i<nItemsToAdd; i++)
  {
    const CStabilityTestItem item{rand() % 20, -1};
    arr5.Add(item);
  }
  _tprintf(_T("array prior to first sort\n  "));
  for (int i{0}; i<nItemsToAdd; i++)
  {
  #ifdef _WIN64
    _tprintf(_T("%d,%I64d "), arr5.GetAt(i).m_nValue, arr5.GetAt(i).m_nOriginalIndex);
  #else
    _tprintf(_T("%d,%d "), arr5.GetAt(i).m_nValue, arr5.GetAt(i).m_nOriginalIndex);
  #endif //#ifdef _WIN64
  }
  arr5.Sort();
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr5.IsSorted());
  //Set the original index value
  for (INT_PTR i{0}; i<arr5.GetSize(); i++)
    arr5[i].m_nOriginalIndex = i;
  _tprintf(_T("\narray after Sort\n  "));
  for (int i{0}; i<nItemsToAdd; i++)
  {
  #ifdef _WIN64
    _tprintf(_T("%d,%I64d "), arr5[i].m_nValue, arr5[i].m_nOriginalIndex);
  #else
    _tprintf(_T("%d,%d "), arr5[i].m_nValue, arr5[i].m_nOriginalIndex);
  #endif //#ifdef _WIN64
  }
  arr5.StableSort();
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr5.IsSorted());
  _tprintf(_T("\narray after StableSort\n  "));
  for (int i{0}; i<nItemsToAdd; i++)
  {
  #ifdef _WIN64
    _tprintf(_T("%d,%I64d "), arr5.GetAt(i).m_nValue, arr5.GetAt(i).m_nOriginalIndex);
  #else
    _tprintf(_T("%d,%d "), arr5.GetAt(i).m_nValue, arr5.GetAt(i).m_nOriginalIndex);
  #endif //#ifdef _WIN64
  }

  //Now validate the stability of the sort
  for (INT_PTR i{0}; i<arr5.GetSize(); i++)
#pragma warning(suppress: 26477)
    ATLASSERT(arr5[i].m_nOriginalIndex == i);

  //Lets test the correctness of the StableSort method.
  CSortedArrayEx<CStabilityTestItem, CompareCStabilityTestItemsClass, const CStabilityTestItem&> arr7;
  _tprintf(_T("\nTesting StableSort correctness"));
  nItemsToAdd = 100;
  arr7.SetSize(0, nItemsToAdd);
  for (int i{0}; i<nItemsToAdd; i++)
  {
    const CStabilityTestItem item{rand() % 20, -1};
    arr7.Add(item);
  }
  //Set the original index value
  for (INT_PTR i{0}; i<arr7.GetSize(); i++)
    arr7[i].m_nOriginalIndex = i;
  _tprintf(_T("\narray prior to StableSort\n  "));
  for (int i{0}; i<nItemsToAdd; i++)
  {
  #ifdef _WIN64
    _tprintf(_T("%d,%I64d "), arr7.GetAt(i).m_nValue, arr7.GetAt(i).m_nOriginalIndex);
  #else
    _tprintf(_T("%d,%d "), arr7.GetAt(i).m_nValue, arr7.GetAt(i).m_nOriginalIndex);
  #endif //#ifdef _WIN64
  }
  arr7.StableSort();
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr7.IsSorted());
  _tprintf(_T("\narray after StableSort\n  "));
  for (int i{0}; i<nItemsToAdd; i++)
  {
  #ifdef _WIN64
    _tprintf(_T("%d,%I64d "), arr7.GetAt(i).m_nValue, arr7.GetAt(i).m_nOriginalIndex);
  #else
    _tprintf(_T("%d,%d "), arr7.GetAt(i).m_nValue, arr7.GetAt(i).m_nOriginalIndex);
  #endif //#ifdef _WIN64
  }
  

  //Test out the OrderedInsert (LeaveIfDuplicate) method
  CSortedArray<int, const int&> idata2;
  idata2.SetCompareFunction(CompareInts);
  d = 7;
  nIndex2 = idata2.FindOrderedInsertIndex(d, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex = idata2.OrderedInsert(d, 1, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 3;
  nIndex2 = idata2.FindOrderedInsertIndex(d, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex = idata2.OrderedInsert(d, 1, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 4;
  nIndex2 = idata2.FindOrderedInsertIndex(d, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex = idata2.OrderedInsert(d, 1, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 5;
  nIndex2 = idata2.FindOrderedInsertIndex(d, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex = idata2.OrderedInsert(d, 1, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 2;
  nIndex2 = idata2.FindOrderedInsertIndex(d, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex = idata2.OrderedInsert(d, 1, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 3;
  nIndex2 = idata2.FindOrderedInsertIndex(d, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex = idata2.OrderedInsert(d, 1, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 4;
  nIndex2 = idata2.FindOrderedInsertIndex(d, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex = idata2.OrderedInsert(d, 1, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 5;
  nIndex2 = idata2.FindOrderedInsertIndex(d, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
  nIndex = idata2.OrderedInsert(d, 1, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate); //NOLINT(clang-analyzer-deadcode.DeadStores)
#pragma warning(suppress: 26477)
  ATLASSERT(nIndex2 == nIndex);
  d = 5; //NOLINT(clang-analyzer-deadcode.DeadStores)

#pragma warning(suppress: 26477)
  ATLASSERT(idata2.FindOrderedInsertIndex(d, CSortedArrayEnums::InsertionBehaviour::FailIfDuplicate) == -1);
#pragma warning(suppress: 26477)
  ATLASSERT(idata2.OrderedInsert(d, 1, CSortedArrayEnums::InsertionBehaviour::FailIfDuplicate) == -1);

#pragma warning(suppress: 26477)
  ATLASSERT(idata2.FindOrderedInsertIndex(d, CSortedArrayEnums::InsertionBehaviour::OverwriteIfDuplicate) == 3);
#pragma warning(suppress: 26477)
  ATLASSERT(idata2.OrderedInsert(d, 1, CSortedArrayEnums::InsertionBehaviour::OverwriteIfDuplicate) == 3);

#pragma warning(suppress: 26477)
  ATLASSERT(idata2.FindOrderedInsertIndex(d, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate) == 3);
#pragma warning(suppress: 26477)
  ATLASSERT(idata2.OrderedInsert(d, 1, CSortedArrayEnums::InsertionBehaviour::LeaveIfDuplicate) == 3);

#pragma warning(suppress: 26477)
  ATLASSERT(idata2.GetSize() == 5);
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(idata2.IsSorted());

  _tprintf(_T("\nItems in array which should have unique items\n  "));
  for (INT_PTR i{0}; i<idata2.GetSize(); i++)
    _tprintf(_T("%d "), idata2[i]);



  //Test performance of function pointer vs functor versions
  CSortedArray<int, const int&> arr1;
  arr1.SetCompareFunction(CompareInts);

  CSortedArrayEx<int, CompareIntsClass, const int&> arr2;

  nItemsToAdd = 1000;
  int nArraysToCreate{100};
  LARGE_INTEGER c1{};
  LARGE_INTEGER c2{};
  LARGE_INTEGER c3{};
  LARGE_INTEGER c4{};
  LARGE_INTEGER cf{};
  QueryPerformanceFrequency(&cf);
  __int64 x1{0};
  __int64 x2{0};
  __int64 xx{0};
  xx = cf.QuadPart / 1000;

  _tprintf(_T("\nTesting sorting speed %d times\n"), nArraysToCreate);

  _tprintf(_T("..Inserting %d items into arrays\n"), nItemsToAdd);
  arr1.SetSize(0, nItemsToAdd);
  arr2.SetSize(0, nItemsToAdd);
  for (int i{nItemsToAdd}; i>=0; i--)
  {
    arr1.Add(rand() % 1000);
    arr2.Add(rand() % 1000);
  }

  _tprintf(_T("..Sorting function pointer array\n"));
  QueryPerformanceCounter(&c1);
  for (int j{0}; j<nArraysToCreate; j++)
    arr1.Sort();
  QueryPerformanceCounter(&c2);

  _tprintf(_T("..Sorting functor array\n"));
  QueryPerformanceCounter(&c3);
  for (int j{0}; j<nArraysToCreate; j++)
    arr2.Sort();
  QueryPerformanceCounter(&c4);
  x1 = c2.QuadPart - c1.QuadPart;
  x2 = c4.QuadPart - c3.QuadPart;
  x1 /= xx;
  x2 /= xx;
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr1.IsSorted());
#pragma warning(suppress: 26477)
  ATLASSERT(arr2.IsSorted());

  _tprintf(_T("..Time to sort function pointer array of size %d: %I64d ms\n"), nItemsToAdd, x1);
  _tprintf(_T("..Time to sort functor array of size %d: %I64d ms\n"), nItemsToAdd, x2);

  _tprintf(_T("..Sample Function pointer array elements --------------\n"));
  for (int i{0}; i<30; i++)
    _tprintf(_T("%d "), arr1[i]);
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr1.IsSorted());

  _tprintf(_T("\n..Sample Functor array elements -----------------------\n"));
  for (int i{0}; i<30; i++)
    _tprintf(_T("%d "), arr2[i]);
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr2.IsSorted());


  nItemsToAdd = 1000;
  nArraysToCreate = 1000;
  _tprintf(_T("\nTesting stable sorting speed %d times\n"), nArraysToCreate);

  _tprintf(_T("..Stable Sorting function pointer array\n"));
  QueryPerformanceCounter(&c1);
  for (int j{0}; j<nArraysToCreate; j++)
  {
    arr1.SetSize(0, nItemsToAdd);
    for (int i{nItemsToAdd}; i>=0; i--)
      arr1.Add(rand() % 1000);
    arr1.StableSort();

    arr2.SetSize(0, nItemsToAdd);
    for (int i{nItemsToAdd}; i>=0; i--)
      arr2.Add(rand() % 1000);

    arr2.StableSort();
  }
  QueryPerformanceCounter(&c2);

  _tprintf(_T("..Stable Sorting functor array\n"));
  QueryPerformanceCounter(&c3);
  for (int j{0}; j<nArraysToCreate; j++)
  {
    arr2.SetSize(0, nItemsToAdd);
    for (int i{nItemsToAdd}; i>=0; i--)
      arr2.Add(rand() % 1000);
    arr2.StableSort();
  }
  QueryPerformanceCounter(&c4);
  x1 = c2.QuadPart - c1.QuadPart; 
  x2 = c4.QuadPart - c3.QuadPart; 
  x1 /= xx;
  x2 /= xx;
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr1.IsSorted());
#pragma warning(suppress: 26477)
  ATLASSERT(arr2.IsSorted());

  _tprintf(_T("..Time to Stable Sort function pointer array of size %d: %I64d ms\n"), nItemsToAdd, x1);
  _tprintf(_T("..Time to Stable Sort functor array of size %d: %I64d ms\n"), nItemsToAdd, x2);

  _tprintf(_T("..Sample Function pointer array elements --------------\n"));
  for (int i{0}; i<30; i++)
    _tprintf(_T("%d "), arr1[i]);
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr1.IsSorted());

  _tprintf(_T("\n..Sample Functor array elements -----------------------\n"));
  for (int i{0}; i<30; i++)
    _tprintf(_T("%d "), arr2[i]);
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr2.IsSorted());


  CSortedArray<int, const int&> arr3;
  arr3.SetCompareFunction(CompareInts);

  CSortedArrayEx<int, CompareIntsClass, const int&> arr4;
  _tprintf(_T("\nTesting OrderedInsert speed\n"));

  _tprintf(_T("..Inserting %d items into function pointer array (in ordered mode)\n"), nItemsToAdd);
  arr3.SetSize(0, nItemsToAdd);
  QueryPerformanceCounter(&c1);
  for (int i{0}; i<nItemsToAdd; i++)
    arr3.OrderedInsert(i);
  QueryPerformanceCounter(&c2);
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr3.IsSorted());

  _tprintf(_T("..Inserting %d items into functor array (in ordered mode)\n"), nItemsToAdd);
  arr4.SetSize(0, nItemsToAdd);
  QueryPerformanceCounter(&c3);
  for (int i{0}; i<nItemsToAdd; i++)
    arr4.OrderedInsert(i);
  QueryPerformanceCounter(&c4);
  //Validate the data is truly sorted
#pragma warning(suppress: 26477)
  ATLASSERT(arr4.IsSorted());

  x1 = c2.QuadPart - c1.QuadPart;
  x2 = c4.QuadPart - c3.QuadPart;
  x1 /= xx;
  x2 /= xx;

  _tprintf(_T("..Time to ordered insert %d elements into function pointer array: %I64d ms\n"), nItemsToAdd, x1);
  _tprintf(_T("..Time to ordered insert %d elements into functor array: %I64d ms\n"), nItemsToAdd, x2);

  _tprintf(_T("..Sample Function pointer array elements --------------\n"));

  for (int i{0}; i<30; i++)
    _tprintf(_T("%d "), arr3[i]);

  _tprintf(_T("\n..Sample Functor array elements ----------------------\n"));
  
  for (int i{0}; i<30; i++)
    _tprintf(_T("%d "), arr4[i]);

  return 0;
}
